/*
 File main.cpp
 
 ** DESCRIPTION **
 PortOpt: a C++ program implementing the Markowitz(1952) mean-variance model
 with agent's linear indifference curves toward risk in order to find the
 optimal assets portfolio under risk.
 
 ** USAGE **
 Run this program from the command line ("start"-> "cmd" -> cd /path/to/portopt
 in windows or a shell terminal in linux) passing as argument the fileneme of
 the assets variance/covariance matrix, the filename of the assets means
 returns and the preferred risk aversion coefficient.

 Run ./portopt --help for all the options or read the README.txt for
 instructions on how to compile.

 In order to minimise the variance this program internally uses QuadProg++, a
 library that implement the algorithm of Goldfarb and Idnani for the solution
 of a (convex) Quadratic Programming problem by means of an active-set dual
 method.
 
 ** LICENSE **
 This file is part of PortOpt: a C++ program implementing the Markowitz(1952)
 mean-variance model with agent's linear indifference curves toward risk
 in order to find the optimal assets portfolio under risk.
 Copyright (C) 2014 Antonello Lobianco.

 PortOpt is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 PortOpt is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public License
 along with PortOpt. If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <sstream>
#include <fstream>
#include <iterator>
#include <string>
#include <vector>
#include <algorithm>
#include "anyoption.h"
#include "portopt.h"

using namespace std;

vector<int>  readCsvData(vector< vector<double>>& data, const string& filename, const string& delimiter, const string& decseparator);

int main (int argc, char *argv[]) {
  
  double alpha;
  string iVARFilename;
  string iMEANSFilename;
  string delimiter = ";";
  string decseparator = ".";
  
  cout << endl;
  cout << "*****************************************************************************" << endl;
  cout << "***          !! Welcome to PortOpt - The Portfolio Optimizer !!           ***" << endl;
  cout << "*** For info & doc: http://lobianco.org/antonello/personal:portfolio:home ***" << endl;
  cout << "***                 Compiled on: " << __DATE__ << " - " << __TIME__ << "                   ***" << endl;
  cout << "*****************************************************************************" << endl<<endl;
  cout << "Find the optimal assets portfolio following the Markowitz(1952) mean-variance "<< endl;
  cout << "model with agent's linear indifference curves toward risk." << endl<<endl;
  
    // 1. CREATE AN OBJECT
  AnyOption *opt = new AnyOption();


  // 2. SET PREFERENCES
  //opt->setVerbose(); // print warnings about unknown options
  //opt->autoUsagePrint(true); // print usage for bad options

  // 3. SET THE USAGE/HELP
  opt->addUsage( "*** PortOpt - The Portfolio Optimizer ***" );
  opt->addUsage( "Usage: " );
  opt->addUsage( "" );
  opt->addUsage( " -h  --help                                   Prints this help " );
  opt->addUsage( " -v  --var-file [input_var_file_name]         Input file containing the variance/covariance matrix (relative path) " );  
  opt->addUsage( " -m  --means-file [input_means_file_name]     Input file containing the means vector (relative path) " ); 
  opt->addUsage( " -a  --alpha [alpha_coefficient]              The coefficient between production and risk in the linear indifference curves  " ); 
  opt->addUsage( " -f  --field-delimiter [field_delimiter]      Character to use as field delimiter (default: ';') " ); 
  opt->addUsage( " -s  --decimal-separator [decimal-separator]  Character to use as decimal delimiter (default: '.') " ); 
  
  opt->addUsage( "" );
  opt->addUsage( "Notes:");
  opt->addUsage( "  - Higher the alpha, lower the agent risk aversion;" );
  opt->addUsage( "  - Set a negative alpha to retrieve the portfolio with the lowest possible variance;" );
  opt->addUsage( "  - Set alpha to zero to retrieve the portfolio with the highest mean, indipendently from variance (solution not guaranteed to be unique)." );
  opt->addUsage( "" );
  opt->addUsage( "Read installed documentation or browse it at http://lobianco.org/antonello/personal:portfolio:home." );
  opt->addUsage( "" );

  // 4. SET THE OPTION STRINGS/CHARACTERS
  opt->setFlag(  "help", 'h' );
  opt->setOption(  "var-file", 'v' );
  opt->setOption(  "means-file", 'm' );
  opt->setOption(  "alpha", 'a' );
  opt->setOption(  "field-delimiter", 'f' );
  opt->setOption(  "decimal-separator", 's' );

  // 5. PROCESS THE COMMANDLINE
  opt->processCommandArgs( argc, argv );

  // 6. GET THE VALUES
  if( opt->getFlag( "help" ) || opt->getFlag( 'h' ) || opt->getArgc() >0 ) {
    opt->printUsage();
    delete opt;
    return EXIT_FAILURE;
  }
  
 
  if( opt->getValue( 'v' ) != NULL  || opt->getValue( "var-file" ) != NULL  ){
      iVARFilename = opt->getValue( 'v' );
  }
  else {
    opt->printUsage();
    delete opt;
    return EXIT_FAILURE;
  }

  if( opt->getValue( 'm' ) != NULL  || opt->getValue( "means-file" ) != NULL  ){
      iMEANSFilename = opt->getValue( 'm' );
  }
  else {
    opt->printUsage();
    delete opt;
    return EXIT_FAILURE;
  }
  
  if( opt->getValue( 'a' ) != NULL  || opt->getValue( "alpha" ) != NULL  ){
    string temp = opt->getValue( 'a' );
    istringstream tempstream(temp);
    tempstream >> alpha;
  }
  else { 
    opt->printUsage();
    delete opt;
    return EXIT_FAILURE;
  }
  
  if( opt->getValue( 'f' ) != NULL  || opt->getValue( "field-delimiter" ) != NULL  ){
      delimiter = opt->getValue( 'f' ); // default already defined
  }
  
  if( opt->getValue( 's' ) != NULL  || opt->getValue( "decimal-separator" ) != NULL  ){
      decseparator = opt->getValue( 's' ); // default already defined
  }
  
  delete opt;
  double o;
  vector<double> MEANS;
  vector<vector<double>>means_temp; 
  double opt_var = 0;
  double opt_mean = 0;
  vector<int> retCodes = {0,0,0};
  vector< vector<double>> VAR;
  int errorcode;
 
  // reading covariance matrix
  retCodes = readCsvData(VAR, iVARFilename, delimiter, decseparator); 
  if(retCodes[0] >0){
    cout << "***ERROR: There has been problem reading the variance file. Sure the name is correct and you have access to the file ?"<<endl; 
    return EXIT_FAILURE; 
  }

  int n = VAR.size();
  vector<double> x(n,0.0); // initialize a vector of n zeros.
  
  // reading means vector 
  retCodes = readCsvData(means_temp, iMEANSFilename, delimiter, decseparator); 
  if(retCodes[0] >0){
  cout << "***ERROR: There has been a problem reading the means file. Sure the name is correct and you have access to the file ?"<<endl; 
  return EXIT_FAILURE; 
  }
  if(retCodes[2] != 1){
    cout << "***ERROR: The means file must be a single-dimension vector, not a matrix !"<<endl; 
    return EXIT_FAILURE;  
  }
  for(int i=0;i<means_temp.size();i++){
    MEANS.push_back(means_temp.at(i).at(0));
  }

 
  o = solveport (VAR, MEANS, alpha, x, errorcode);
  
  if(errorcode){
    if(errorcode == 1){
      cout << "***ERROR: An error happened in the input data !" << endl;
    } else if (errorcode == 2) {
      cout << "***ERROR: The problem has no solutions" << endl; 
    } else {
      cout << "***ERROR: An unknow error happened (look on the code of portopt.cpp)" << endl; 
    }
    return EXIT_FAILURE; 
  }  
  for (int i = 0; i < x.size(); i++){
    opt_mean += x.at(i)* MEANS.at(i);
    for (int j = 0; j < x.size(); j++) {
      opt_var += x.at(i)*x.at(j)* VAR.at(i).at(j);
    }
  }
  cout << "Optimal portfolio variance: " << opt_var << endl;
  cout << "Optimal portfolio mean: " << opt_mean << endl;
  cout << "Cost of the optimisation function: " << o << endl;
  cout << "Asset shares: " << endl;
  for (int i = 0; i < x.size(); i++){
    cout << i << ": " << x[i] << endl;
  } 
}

// ############## auxiliary functions... #########################


/**
 * Parse a CSV data file and fill the 2d STL vector "data".
 * Limits: only "pure datas" of doubles, not encapsulated by " and without \n inside.
 * Further no formatting in the data (e.g. scientific notation)
 * It however handles both dots and commas as decimal separators and removes thousand separator.
 * 
 * returnCodes[0]: file access 0-> ok 1-> not able to read; 2-> decimal separator equal to comma separator
 * returnCodes[1]: number of records
 * returnCodes[2]: number of fields. -1 If rows have different field size
 * 
 */
vector<int>
readCsvData (vector <vector <double>>& data, const string& filename, const string& delimiter, const string& decseparator){
 
 int vv[3] = { 0,0,0 };
 vector<int> returnCodes(&vv[0], &vv[0]+3);

 string rowstring, stringtoken;
 double doubletoken;
 int rowcount=0;
 int fieldcount=0;
 data.clear();
 
 ifstream iFile(filename, ios_base::in);
 if (!iFile.is_open()){
   returnCodes[0] = 1;
   return returnCodes;
 }
 while (getline(iFile, rowstring)) {
    if (rowstring=="") continue; // empty line
    rowcount ++; //let's start with 1
    if(delimiter == decseparator){
      returnCodes[0] = 2;
      return returnCodes;
    }
    if(decseparator != "."){
     // remove dots (used as thousand separators)
     string::iterator end_pos = remove(rowstring.begin(), rowstring.end(), '.');
     rowstring.erase(end_pos, rowstring.end());
     // replace decimal separator with dots.
     replace(rowstring.begin(), rowstring.end(),decseparator.c_str()[0], '.'); 
    } else {
     // remove commas (used as thousand separators)
     string::iterator end_pos = remove(rowstring.begin(), rowstring.end(), ',');
     rowstring.erase(end_pos, rowstring.end());
    }
    // tokenize..
    vector<double> tokens;
    // Skip delimiters at beginning.
    string::size_type lastPos = rowstring.find_first_not_of(delimiter, 0);
    // Find first "non-delimiter".
    string::size_type pos     = rowstring.find_first_of(delimiter, lastPos);
    while (string::npos != pos || string::npos != lastPos){
        // Found a token, convert it to double add it to the vector.
        stringtoken = rowstring.substr(lastPos, pos - lastPos);
        if (stringtoken == "") {
	  tokens.push_back(0.0);
	} else {
          istringstream totalSString(stringtoken);
	  totalSString >> doubletoken;
	  tokens.push_back(doubletoken);
	}     
        // Skip delimiters.  Note the "not_of"
        lastPos = rowstring.find_first_not_of(delimiter, pos);
        // Find next "non-delimiter"
        pos = rowstring.find_first_of(delimiter, lastPos);
    }
    if(rowcount == 1){
      fieldcount = tokens.size();
      returnCodes[2] = tokens.size();
    } else {
      if ( tokens.size() != fieldcount){
	returnCodes[2] = -1;
      }
    }
    data.push_back(tokens);
 }
 iFile.close();
 returnCodes[1] = rowcount;
 return returnCodes;
}
// --------------------------------------------------------------------------------



