/*
 File portopt.cpp
 
 ** DESCRIPTION **
 PortOpt: a C++ program implementing the Markowitz(1952) mean-variance model
 with agent's linear indifference curves toward risk in order to find the
 optimal assets portfolio under risk.
 
 ** USAGE **
 Run this program from the command line ("start"-> "cmd" -> cd /path/to/portopt
 in windows or a shell terminal in linux) passing as argument the fileneme of
 the assets variance/covariance matrix, the filename of the assets means
 returns and the preferred risk aversion coefficient.

 Run ./portopt --help for all the options or read the README.txt for
 instructions on how to compile.

 In order to minimise the variance this program internally uses QuadProg++, a
 library that implement the algorithm of Goldfarb and Idnani for the solution
 of a (convex) Quadratic Programming problem by means of an active-set dual
 method.
 
 ** LICENSE **
 This file is part of PortOpt: a C++ program implementing the Markowitz(1952)
 mean-variance model with agent's linear indifference curves toward risk
 in order to find the optimal assets portfolio under risk.
 Copyright (C) 2014 Antonello Lobianco.

 PortOpt is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 PortOpt is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public License
 along with PortOpt. If not, see <http://www.gnu.org/licenses/>.
*/


#include <iostream>
#include <sstream>
#include <fstream>
#include <iterator>
#include <string>
#include <vector>
#include <algorithm>
#include <limits>
#include "portopt.h"
#include "QuadProg++.h"

using namespace std;


double solveport (const vector< vector <double> > &VAR, const vector<double> &MEANS, const double &alpha, vector<double> &x_h, int & errorcode ) {
  QuadProgPP::Matrix<double> G, CE, CI;
  QuadProgPP::Vector<double> g0, ce0, ci0, x;
  int n, m, p; // n = var number, m = equality constrains; p = inequality constrains
  double f;
  double alpha_coeff=1;
  errorcode = 0; // 0 -> ok; 1 -> input error; 2 -> no solutions
  // in python swig interface x_h is empty
  
  n = MEANS.size();
  
  bool input_isok=false;
  input_isok = i_checks(VAR,MEANS);
  
  if(!input_isok){
    //cout << "***ERROR: An error happened in the input data !"<<endl; // --> let the client to deal with errors 
    errorcode = 1;
    return 0;
  }
  
  vector<double> x_temp(n,0.0); // initialize a vector of n zeros.
  x_h = x_temp;


  if(alpha>=0){
    alpha_coeff = alpha;
  }

  // objective quadratic matrix
  G.resize(n, n);
   for (int i = 0; i < n; i++) {
    for (int j = 0; j < n; j++) {
      G[i][j] = 2*alpha_coeff*VAR.at(i).at(j);
    }
  }
 // objective linear vector
 g0.resize(n);
  if(alpha>=0){
    for (int i = 0; i < n; i++) {
      g0[i] = -MEANS[i];
    }
  } else {
    for (int i = 0; i < n; i++) {
      g0[i] = 0;
    }
  }
  // equality constrains (sum_x = 1)
  m = 1;
  CE.resize(n, m);
  for (int i = 0; i < n; i++) {
    CE[i][0] = 1;
  }
  ce0.resize(m);
  ce0[0] = -1;
  // inequality constrains (x_i >= 0 foreach i)
  p = n;
  CI.resize(n, p);
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < p; j++) {
      if (i==j){
        CI[i][j] = 1;
      } else {
  CI[i][j] = 0;
      }
    }
  }
  ci0.resize(p);
  for (int i = 0; i < p; i++) {
    ci0[i] = 0;
  }
  // ok, problem is now ready to be optimised by quadprog++..
  if(alpha==0){
    //not really an optimisation....
    int maxMEANScounter = 0;
    double maxMEANS = -numeric_limits<float>::max();
    double minVAR = numeric_limits<float>::max();
    for(int i =0;i<n;i++){
      if (MEANS[i] > maxMEANS) {
        maxMEANS = MEANS[i];
        maxMEANScounter = i;
        minVAR = VAR[i][i];
      } else if (MEANS[i] == maxMEANS ) {
        if (VAR[i][i] < minVAR){
          maxMEANS = MEANS[i];
          maxMEANScounter = i;
          minVAR = VAR[i][i];
        }

      }
    }
    for (int i = 0; i < n; i++) {
      if (i == maxMEANScounter){
        x_h[i] = 1;
      } else {
        x_h[i] = 0;
      }
    }
    f = - MEANS[maxMEANScounter];
  } else {
    x.resize(n);
    f = solve_quadprog(G, g0, CE, ce0, CI, ci0, x);  // <<<<< OPTIMISATION ROUTINE
    if(f == std::numeric_limits<double>::infinity()){
      errorcode=2;
      return 0;
    }
    for (int i = 0; i < n; i++) {
      x_h[i] = x[i];
    }
  }
  return f;
}


bool i_checks(const vector< vector <double> > &VAR, const vector<double> &MEANS){
  
  int n = MEANS.size();
  
  if(VAR.size() != n){
    cout << "***ERROR: The number of variables in the means vector is not equal to the number of variables in the covariance matrix!"<<endl; 
    return false; 
    // todo: move here the other errors checks, as this is a lib now
  }
  
  
  for(int i=0;i<n;i++){
    if(VAR[i].size() != n){
      cout << "***ERROR: The variance matrix is not squared !"<<endl; 
      return false;
    }
    if(MEANS[i] < 0){
      cout << "***ERROR: Only positive values in the means vector are supported"<<endl; 
      return false;
    }
    for(int j=0;j<n;j++){
      if(i==j){
        if(VAR[i][j]<0){
          cout << "***ERROR: The covariance matrix has negative values on the diagonal" << endl;
          return false;
        }
      } else {
        if(VAR[i][j] != VAR[j][i]){
          cout << "***ERROR: The covariance matrix is not simmetric" << endl;
          return false;
        }
      }
    }
  }  
}